@@ -6,19 +6,20 @@ For the documentation, refer to :mod:`ssl` module manual.
 
 This module implements cooperative SSL socket wrappers.
 """
-# Our import magic sadly makes this warning useless
-# pylint: disable=undefined-variable
 
 from __future__ import absolute_import
 import ssl as __ssl__
 
-_ssl = __ssl__._ssl # pylint:disable=no-member
+_ssl = __ssl__._ssl
 
 import errno
+import sys
+
 from gevent.socket import socket, timeout_default
 from gevent.socket import error as socket_error
 from gevent.socket import timeout as _socket_timeout
 from gevent._util import copy_globals
+from gevent._compat import PY36
 
 from weakref import ref as _wref
 
@@ -29,12 +30,38 @@ __implements__ = [
     'get_server_certificate',
 ]
 
+# Manually import things we use so we get better linting.
+# Also, in the past (adding 3.9 support) it turned out we were
+# relying on certain global variables being defined in the ssl module
+# that weren't required to be there, e.g., AF_INET, which should be imported
+# from socket
+from socket import AF_INET
+from socket import SOCK_STREAM
+from socket import SO_TYPE
+from socket import SOL_SOCKET
+
+from ssl import SSLWantReadError
+from ssl import SSLWantWriteError
+from ssl import CERT_NONE
+from ssl import SSLError
+from ssl import SSL_ERROR_EOF
+from ssl import SSL_ERROR_WANT_READ
+from ssl import SSL_ERROR_WANT_WRITE
+from ssl import PROTOCOL_SSLv23
+from ssl import SSLObject
+from ssl import match_hostname
+from ssl import CHANNEL_BINDING_TYPES
+from ssl import CERT_REQUIRED
+from ssl import DER_cert_to_PEM_cert
+from ssl import create_connection
+
 # Import all symbols from Python's ssl.py, except those that we are implementing
 # and "private" symbols.
-__imports__ = copy_globals(__ssl__, globals(),
-                           # SSLSocket *must* subclass gevent.socket.socket; see issue 597
-                           names_to_ignore=__implements__ + ['socket'],
-                           dunder_names_to_keep=())
+__imports__ = copy_globals(
+    __ssl__, globals(),
+    # SSLSocket *must* subclass gevent.socket.socket; see issue 597
+    names_to_ignore=__implements__ + ['socket'],
+    dunder_names_to_keep=())
 
 __all__ = __implements__ + __imports__
 if 'namedtuple' in __all__:
@@ -42,9 +69,36 @@ if 'namedtuple' in __all__:
 
 orig_SSLContext = __ssl__.SSLContext # pylint:disable=no-member
 
+# We have to pass the raw stdlib socket to SSLContext.wrap_socket.
+# That method in turn can pass that object on to things like SNI callbacks.
+# It wouldn't have access to any of the attributes on the SSLSocket, like
+# context, that it's supposed to (see test_ssl.test_sni_callback). Previously
+# we just delegated to the sslsocket with __getattr__, but 3.8
+# added some new callbacks and a test that the object they get is an instance
+# of the high-level SSLSocket class, so that doesn't work anymore. Instead,
+# we wrap the callback and get the real socket to pass on.
+class _contextawaresock(socket._gevent_sock_class):
+    __slots__ = ('_sslsock',)
+
+    def __init__(self, family, type, proto, fileno, sslsocket_wref):
+        super().__init__(family, type, proto, fileno)
+        self._sslsock = sslsocket_wref
+
+class _Callback(object):
+
+    __slots__ = ('user_function',)
+
+    def __init__(self, user_function):
+        self.user_function = user_function
+
+    def __call__(self, conn, *args):
+        conn = conn._sslsock()
+        return self.user_function(conn, *args)
 
 class SSLContext(orig_SSLContext):
 
+    __slots__ = ()
+
     # Added in Python 3.7
     sslsocket_class = None # SSLSocket is assigned later
 
@@ -53,7 +107,7 @@ class SSLContext(orig_SSLContext):
                     suppress_ragged_eofs=True,
                     server_hostname=None,
                     session=None):
-        # pylint:disable=arguments-differ
+        # pylint:disable=arguments-differ,not-callable
         # (3.6 adds session)
         # Sadly, using *args and **kwargs doesn't work
         return self.sslsocket_class(
@@ -64,10 +118,6 @@ class SSLContext(orig_SSLContext):
             _context=self,
             _session=session)
 
-    if not hasattr(orig_SSLContext, 'check_hostname'):
-        # Python 3.3 lacks this
-        check_hostname = False
-
     if hasattr(orig_SSLContext.options, 'setter'):
         # In 3.6, these became properties. They want to access the
         # property __set__ method in the superclass, and they do so by using
@@ -89,6 +139,7 @@ class SSLContext(orig_SSLContext):
 
     if hasattr(orig_SSLContext, 'minimum_version'):
         # Like the above, added in 3.7
+        # pylint:disable=no-member
         @orig_SSLContext.minimum_version.setter
         def minimum_version(self, value):
             super(orig_SSLContext, orig_SSLContext).minimum_version.__set__(self, value)
@@ -97,74 +148,61 @@ class SSLContext(orig_SSLContext):
         def maximum_version(self, value):
             super(orig_SSLContext, orig_SSLContext).maximum_version.__set__(self, value)
 
+    if hasattr(orig_SSLContext, '_msg_callback'):
+        # And ditto for 3.8
+        # msg_callback is more complex because they want to actually *do* stuff
+        # in the setter, so we need to call it. For that to work we temporarily rebind
+        # SSLContext back. This function cannot switch, so it should be safe,
+        # unless somehow we have multiple threads in a monkey-patched ssl module
+        # at the same time, which doesn't make much sense.
+        @property
+        def _msg_callback(self):
+            result = super()._msg_callback
+            if isinstance(result, _Callback):
+                result = result.user_function
+            return result
 
-class _contextawaresock(socket._gevent_sock_class): # Python 2: pylint:disable=slots-on-old-class
-    # We have to pass the raw stdlib socket to SSLContext.wrap_socket.
-    # That method in turn can pass that object on to things like SNI callbacks.
-    # It wouldn't have access to any of the attributes on the SSLSocket, like
-    # context, that it's supposed to (see test_ssl.test_sni_callback). Our
-    # solution is to keep a weak reference to the SSLSocket on the raw
-    # socket and delegate.
+        @_msg_callback.setter
+        def _msg_callback(self, value):
+            if value and callable(value):
+                value = _Callback(value)
 
-    # We keep it in a slot to avoid having the ability to set any attributes
-    # we're not prepared for (because we don't know what to delegate.)
+            __ssl__.SSLContext = orig_SSLContext
+            try:
+                super(SSLContext, SSLContext)._msg_callback.__set__(self, value) # pylint:disable=no-member
+            finally:
+                __ssl__.SSLContext = SSLContext
 
-    __slots__ = ('_sslsock',)
+    if hasattr(orig_SSLContext, 'sni_callback'):
+        # Added in 3.7.
+        @property
+        def sni_callback(self):
+            result = super().sni_callback
+            if isinstance(result, _Callback):
+                result = result.user_function # pylint:disable=no-member
+            return result
+        @sni_callback.setter
+        def sni_callback(self, value):
+            if value and callable(value):
+                value = _Callback(value)
+            super(orig_SSLContext, orig_SSLContext).sni_callback.__set__(self, value)
+    else:
+        # In newer versions, this just sets sni_callback.
+        def set_servername_callback(self, cb): # pylint:disable=arguments-differ
+            if cb and callable(cb):
+                cb = _Callback(cb)
+            super().set_servername_callback(cb)
 
-    @property
-    def context(self):
-        return self._sslsock().context
-
-    @context.setter
-    def context(self, ctx):
-        self._sslsock().context = ctx
-
-    @property
-    def session(self):
-        """The SSLSession for client socket."""
-        return self._sslsock().session
-
-    @session.setter
-    def session(self, session):
-        self._sslsock().session = session
-
-    def __getattr__(self, name):
-        try:
-            return getattr(self._sslsock(), name)
-        except RuntimeError:
-            # XXX: If the attribute doesn't exist,
-            # we infinitely recurse
-            pass
-        raise AttributeError(name)
-
-try:
-    _SSLObject_factory = SSLObject
-except NameError:
-    # 3.4 and below do not have SSLObject, something
-    # we magically import through copy_globals
-    pass
-else:
-    if hasattr(SSLObject, '_create'):
-        # 3.7 is making thing difficult and won't let you
-        # actually construct an object
-        def _SSLObject_factory(sslobj, owner=None, session=None):
-            s = SSLObject.__new__(SSLObject)
-            s._sslobj = sslobj
-            s._sslobj.owner = owner or s
-            if session is not None:
-                s._sslobj.session = session
-            return s
 
 class SSLSocket(socket):
     """
-    gevent `ssl.SSLSocket <https://docs.python.org/3/library/ssl.html#ssl-sockets>`_
-    for Python 3.
+    gevent `ssl.SSLSocket
+    <https://docs.python.org/3/library/ssl.html#ssl-sockets>`_ for
+    Python 3.
     """
 
     # pylint:disable=too-many-instance-attributes,too-many-public-methods
 
-    _gevent_sock_class = _contextawaresock
-
     def __init__(self, sock=None, keyfile=None, certfile=None,
                  server_side=False, cert_reqs=CERT_NONE,
                  ssl_version=PROTOCOL_SSLv23, ca_certs=None,
@@ -174,8 +212,11 @@ class SSLSocket(socket):
                  server_hostname=None,
                  _session=None, # 3.6
                  _context=None):
+        # When a *sock* argument is passed, it is used only for its fileno()
+        # and is immediately detach()'d *unless* we raise an error.
 
         # pylint:disable=too-many-locals,too-many-statements,too-many-branches
+
         if _context:
             self._context = _context
         else:
@@ -222,38 +263,42 @@ class SSLSocket(socket):
         self.suppress_ragged_eofs = suppress_ragged_eofs
         connected = False
         if sock is not None:
+            timeout = sock.gettimeout()
             socket.__init__(self,
                             family=sock.family,
                             type=sock.type,
                             proto=sock.proto,
                             fileno=sock.fileno())
-            self.settimeout(sock.gettimeout())
-            # see if it's connected
-            try:
-                sock.getpeername()
-            except socket_error as e:
-                if e.errno != errno.ENOTCONN:
-                    raise
-            else:
-                connected = True
+            self.settimeout(timeout)
+            # When Python 3 sockets are __del__, they close() themselves,
+            # including their underlying fd, unless they have been detached.
+            # Only detach if we succeed in taking ownership; if we raise an exception,
+            # then the user might have no way to close us and release the resources.
             sock.detach()
         elif fileno is not None:
             socket.__init__(self, fileno=fileno)
         else:
             socket.__init__(self, family=family, type=type, proto=proto)
 
-        self._sock._sslsock = _wref(self)
         self._closed = False
         self._sslobj = None
+        # see if we're connected
+        try:
+            self._sock.getpeername()
+        except socket_error as e:
+            if e.errno != errno.ENOTCONN:
+                # This file descriptor is hosed, shared or not.
+                # Clean up.
+                self.close()
+                raise
+        else:
+            connected = True
         self._connected = connected
         if connected:
             # create the SSL object
             try:
-                self._sslobj = self._context._wrap_socket(self._sock, server_side,
-                                                          server_hostname)
-                if _session is not None: # 3.6+
-                    self._sslobj = _SSLObject_factory(self._sslobj,  owner=self,
-                                                      session=self._session)
+                self._sslobj = self.__create_sslobj(server_side, _session)
+
                 if do_handshake_on_connect:
                     timeout = self.gettimeout()
                     if timeout == 0.0:
@@ -265,6 +310,16 @@ class SSLSocket(socket):
                 self.close()
                 raise x
 
+    def _gevent_sock_class(self, family, type, proto, fileno):
+        return _contextawaresock(family, type, proto, fileno, _wref(self))
+
+    def _extra_repr(self):
+        return ' server=%s, cipher=%r' % (
+            self.server_side,
+            self._sslobj.cipher() if self._sslobj is not None else ''
+
+        )
+
     @property
     def context(self):
         return self._context
@@ -308,23 +363,26 @@ class SSLSocket(socket):
             # EAGAIN.
             self.getpeername()
 
-    def read(self, len=1024, buffer=None):
+    def read(self, nbytes=2014, buffer=None):
         """Read up to LEN bytes and return them.
         Return zero-length string on EOF."""
         # pylint:disable=too-many-branches
         self._checkClosed()
-
+        # The stdlib signature is (len=1024, buffer=None)
+        # but that shadows the len builtin, and its hard/annoying to
+        # get it back.
+        initial_buf_len = len(buffer) if buffer is not None else None
         while True:
             if not self._sslobj:
                 raise ValueError("Read on closed or unwrapped SSL socket.")
-            if len == 0:
+            if nbytes == 0:
                 return b'' if buffer is None else 0
             # Negative lengths are handled natively when the buffer is None
             # to raise a ValueError
             try:
                 if buffer is not None:
-                    return self._sslobj.read(len, buffer)
-                return self._sslobj.read(len or 1024)
+                    return self._sslobj.read(nbytes, buffer)
+                return self._sslobj.read(nbytes or 1024)
             except SSLWantReadError:
                 if self.timeout == 0.0:
                     raise
@@ -336,11 +394,17 @@ class SSLSocket(socket):
                 self._wait(self._write_event, timeout_exc=_SSLErrorReadTimeout)
             except SSLError as ex:
                 if ex.args[0] == SSL_ERROR_EOF and self.suppress_ragged_eofs:
-                    if buffer is None:
-                        return b''
-                    return 0
-                else:
-                    raise
+                    return b'' if buffer is None else len(buffer) - initial_buf_len
+                raise
+            except ConnectionResetError:
+                # Certain versions of Python, built against certain
+                # versions of OpenSSL operating in certain modes,
+                # can produce this instead of SSLError. Notably, it looks
+                # like anything built against 1.1.1c do?
+                if self.suppress_ragged_eofs:
+                    return b'' if buffer is None else len(buffer) - initial_buf_len
+                raise
+
 
     def write(self, data):
         """Write DATA to the underlying SSL channel.  Returns
@@ -450,10 +514,9 @@ class SSLSocket(socket):
         if self._sslobj:
             raise ValueError("sendto not allowed on instances of %s" %
                              self.__class__)
-        elif addr is None:
+        if addr is None:
             return socket.sendto(self, data, flags_or_addr)
-        else:
-            return socket.sendto(self, data, flags_or_addr, addr)
+        return socket.sendto(self, data, flags_or_addr, addr)
 
     def sendmsg(self, *args, **kwargs):
         # Ensure programs don't send data unencrypted if they try to
@@ -489,8 +552,7 @@ class SSLSocket(socket):
                 # Python #23804
                 return b''
             return self.read(buflen)
-        else:
-            return socket.recv(self, buflen, flags)
+        return socket.recv(self, buflen, flags)
 
     def recv_into(self, buffer, nbytes=None, flags=0):
         self._checkClosed()
@@ -502,24 +564,21 @@ class SSLSocket(socket):
             if flags != 0:
                 raise ValueError("non-zero flags not allowed in calls to recv_into() on %s" % self.__class__)
             return self.read(nbytes, buffer)
-        else:
-            return socket.recv_into(self, buffer, nbytes, flags)
+        return socket.recv_into(self, buffer, nbytes, flags)
 
     def recvfrom(self, buflen=1024, flags=0):
         self._checkClosed()
         if self._sslobj:
             raise ValueError("recvfrom not allowed on instances of %s" %
                              self.__class__)
-        else:
-            return socket.recvfrom(self, buflen, flags)
+        return socket.recvfrom(self, buflen, flags)
 
     def recvfrom_into(self, buffer, nbytes=None, flags=0):
         self._checkClosed()
         if self._sslobj:
             raise ValueError("recvfrom_into not allowed on instances of %s" %
                              self.__class__)
-        else:
-            return socket.recvfrom_into(self, buffer, nbytes, flags)
+        return socket.recvfrom_into(self, buffer, nbytes, flags)
 
     def recvmsg(self, *args, **kwargs):
         raise NotImplementedError("recvmsg not allowed on instances of %s" %
@@ -541,44 +600,52 @@ class SSLSocket(socket):
         socket.shutdown(self, how)
 
     def unwrap(self):
-        if self._sslobj:
-            while True:
-                try:
-                    s = self._sslobj.shutdown()
-                    break
-                except SSLWantReadError:
-                    # Callers of this method expect to get a socket
-                    # back, so we can't simply return 0, we have
-                    # to let these be raised
-                    if self.timeout == 0.0:
-                        raise
-                    self._wait(self._read_event)
-                except SSLWantWriteError:
-                    if self.timeout == 0.0:
-                        raise
-                    self._wait(self._write_event)
-
-            self._sslobj = None
-
-            # The return value of shutting down the SSLObject is the
-            # original wrapped socket passed to _wrap_socket, i.e.,
-            # _contextawaresock. But that object doesn't have the
-            # gevent wrapper around it so it can't be used. We have to
-            # wrap it back up with a gevent wrapper.
-            assert s is self._sock
-            # In the stdlib, SSLSocket subclasses socket.socket and passes itself
-            # to _wrap_socket, so it gets itself back. We can't do that, we have to
-            # pass our subclass of _socket.socket, _contextawaresock.
-            # So ultimately we should return ourself.
-
-            # See test_ftplib.py:TestTLS_FTPClass.test_ccc
-            return self
-        else:
+        if not self._sslobj:
             raise ValueError("No SSL wrapper around " + str(self))
 
+        try:
+            # 3.7 and newer, that use the SSLSocket object
+            # call its shutdown.
+            shutdown = self._sslobj.shutdown
+        except AttributeError:
+            # Earlier versions use SSLObject, which covers
+            # that with a layer.
+            shutdown = self._sslobj.unwrap
+
+        while True:
+            try:
+                s = shutdown()
+                break
+            except SSLWantReadError:
+                # Callers of this method expect to get a socket
+                # back, so we can't simply return 0, we have
+                # to let these be raised
+                if self.timeout == 0.0:
+                    raise
+                self._wait(self._read_event)
+            except SSLWantWriteError:
+                if self.timeout == 0.0:
+                    raise
+                self._wait(self._write_event)
+
+        self._sslobj = None
+
+        # The return value of shutting down the SSLObject is the
+        # original wrapped socket passed to _wrap_socket, i.e.,
+        # _contextawaresock. But that object doesn't have the
+        # gevent wrapper around it so it can't be used. We have to
+        # wrap it back up with a gevent wrapper.
+        assert s is self._sock
+        # In the stdlib, SSLSocket subclasses socket.socket and passes itself
+        # to _wrap_socket, so it gets itself back. We can't do that, we have to
+        # pass our subclass of _socket.socket, _contextawaresock.
+        # So ultimately we should return ourself.
+
+        # See test_ftplib.py:TestTLS_FTPClass.test_ccc
+        return self
+
     def _real_close(self):
         self._sslobj = None
-        # self._closed = True
         socket._real_close(self)
 
     def do_handshake(self):
@@ -605,6 +672,25 @@ class SSLSocket(socket):
                                  "argument")
             match_hostname(self.getpeercert(), self.server_hostname)
 
+    if hasattr(SSLObject, '_create'):
+        # 3.7+, making it difficult to create these objects.
+        # There's a new type, _ssl.SSLSocket, that takes the
+        # place of SSLObject for self._sslobj. This one does it all.
+        def __create_sslobj(self, server_side=False, session=None):
+            return self.context._wrap_socket(
+                self._sock, server_side, self.server_hostname,
+                owner=self._sock, session=session
+            )
+    elif PY36: # 3.6
+        def __create_sslobj(self, server_side=False, session=None):
+            sslobj = self._context._wrap_socket(self._sock, server_side, self.server_hostname)
+            return SSLObject(sslobj, owner=self._sock, session=session)
+    else: # 3.5
+        def __create_sslobj(self, server_side=False, session=None): # pylint:disable=unused-argument
+            sslobj = self._context._wrap_socket(self._sock, server_side, self.server_hostname)
+            return SSLObject(sslobj, owner=self._sock)
+
+
     def _real_connect(self, addr, connect_ex):
         if self.server_side:
             raise ValueError("can't connect in server-side mode")
@@ -612,9 +698,8 @@ class SSLSocket(socket):
         # connected at the time of the call.  We connect it, then wrap it.
         if self._connected:
             raise ValueError("attempt to connect already-connected SSLSocket!")
-        self._sslobj = self._context._wrap_socket(self._sock, False, self.server_hostname)
-        if self._session is not None: # 3.6+
-            self._sslobj = _SSLObject_factory(self._sslobj, owner=self, session=self._session)
+        self._sslobj = self.__create_sslobj(False, self._session)
+
         try:
             if connect_ex:
                 rc = socket.connect_ex(self, addr)
@@ -641,17 +726,23 @@ class SSLSocket(socket):
         return self._real_connect(addr, True)
 
     def accept(self):
-        """Accepts a new connection from a remote client, and returns
-        a tuple containing that new connection wrapped with a server-side
-        SSL channel, and the address of the remote client."""
-
-        newsock, addr = socket.accept(self)
-        newsock._drop_events()
-        newsock = self._context.wrap_socket(newsock,
-                                            do_handshake_on_connect=self.do_handshake_on_connect,
-                                            suppress_ragged_eofs=self.suppress_ragged_eofs,
-                                            server_side=True)
-        return newsock, addr
+        """
+        Accepts a new connection from a remote client, and returns a
+        tuple containing that new connection wrapped with a
+        server-side SSL channel, and the address of the remote client.
+        """
+        newsock, addr = super().accept()
+        try:
+            newsock = self._context.wrap_socket(
+                newsock,
+                do_handshake_on_connect=self.do_handshake_on_connect,
+                suppress_ragged_eofs=self.suppress_ragged_eofs,
+                server_side=True
+            )
+            return newsock, addr
+        except:
+            newsock.close()
+            raise
 
     def get_channel_binding(self, cb_type="tls-unique"):
         """Get channel binding data for current connection.  Raise ValueError
@@ -669,6 +760,11 @@ class SSLSocket(socket):
             return None
         return self._sslobj.tls_unique_cb()
 
+    def verify_client_post_handshake(self):
+        # Only present in 3.7.1+; an attributeerror is alright
+        if self._sslobj:
+            return self._sslobj.verify_client_post_handshake()
+        raise ValueError("No SSL wrapper around " + str(self))
 
 # Python does not support forward declaration of types
 SSLContext.sslsocket_class = SSLSocket
@@ -706,9 +802,9 @@ def get_server_certificate(addr, ssl_version=PROTOCOL_SSLv23, ca_certs=None):
         cert_reqs = CERT_REQUIRED
     else:
         cert_reqs = CERT_NONE
-    s = create_connection(addr)
-    s = wrap_socket(s, ssl_version=ssl_version,
-                    cert_reqs=cert_reqs, ca_certs=ca_certs)
-    dercert = s.getpeercert(True)
-    s.close()
+    with create_connection(addr) as sock:
+        with wrap_socket(sock, ssl_version=ssl_version,
+                         cert_reqs=cert_reqs, ca_certs=ca_certs) as sslsock:
+            dercert = sslsock.getpeercert(True)
+    sslsock = sock = None
     return DER_cert_to_PEM_cert(dercert)
